@@ -1,21 +1,28 @@
 """
 PynamoDB exceptions
 """
+import sys
+from dataclasses import dataclass
 from typing import Any
 from typing import Dict
 from typing import Iterable
 from typing import List
 from typing import Optional
+if sys.version_info >= (3, 8):
+    from typing import Literal
+else:
+    from typing_extensions import Literal
 
 import botocore.exceptions
 
 
 class PynamoDBException(Exception):
+    """
+    Base class for all PynamoDB exceptions.
+    """
+
     msg: str
 
-    """
-    A common exception class
-    """
     def __init__(self, msg: Optional[str] = None, cause: Optional[Exception] = None) -> None:
         self.msg = msg if msg is not None else self.msg
         self.cause = cause
@@ -23,10 +30,22 @@ class PynamoDBException(Exception):
 
     @property
     def cause_response_code(self) -> Optional[str]:
+        """
+        The DynamoDB response code such as:
+
+        - ``ConditionalCheckFailedException``
+        - ``ProvisionedThroughputExceededException``
+        - ``TransactionCanceledException``
+
+        Inspect this value to determine the cause of the error and handle it.
+        """
         return getattr(self.cause, 'response', {}).get('Error', {}).get('Code')
 
     @property
     def cause_response_message(self) -> Optional[str]:
+        """
+        The human-readable description of the error returned by DynamoDB.
+        """
         return getattr(self.cause, 'response', {}).get('Error', {}).get('Message')
 
 
@@ -102,21 +121,25 @@ class TableDoesNotExist(PynamoDBException):
         super(TableDoesNotExist, self).__init__(msg)
 
 
+@dataclass
 class CancellationReason:
     """
     A reason for a transaction cancellation.
-    """
-    def __init__(self, *, code: str, message: Optional[str]) -> None:
-        self.code = code
-        self.message = message
+    
+    For a list of possible cancellation reasons and their semantics,
+    see `TransactGetItems`_ and `TransactWriteItems`_ in the AWS documentation.
 
-    def __eq__(self, other: Any) -> bool:
-        return isinstance(other, CancellationReason) and self.code == other.code and self.message == other.message
+    .. _TransactGetItems: https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_TransactGetItems.html
+    .. _TransactWriteItems: https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_TransactWriteItems.html
+    """
+    code: str
+    message: Optional[str] = None
+    raw_item: Optional[Dict[str, Dict[str, Any]]] = None
 
 
 class TransactWriteError(PynamoDBException):
     """
-    Raised when a TransactWrite operation fails
+    Raised when a :class:`~pynamodb.transactions.TransactWrite` operation fails.
     """
 
     @property
@@ -138,7 +161,7 @@ class TransactWriteError(PynamoDBException):
 
 class TransactGetError(PynamoDBException):
     """
-    Raised when a TransactGet operation fails
+    Raised when a :class:`~pynamodb.transactions.TransactGet` operation fails.
     """
     @property
     def cancellation_reasons(self) -> List[Optional[CancellationReason]]:
@@ -159,14 +182,14 @@ class TransactGetError(PynamoDBException):
 
 class InvalidStateError(PynamoDBException):
     """
-    Raises when the internal state of an operation context is invalid
+    Raises when the internal state of an operation context is invalid.
     """
     msg = "Operation in invalid state"
 
 
 class AttributeDeserializationError(TypeError):
     """
-    Raised when attribute type is invalid
+    Raised when attribute type is invalid during deserialization.
     """
     def __init__(self, attr_name: str, attr_type: str):
         msg = "Cannot deserialize '{}' attribute from type: {}".format(attr_name, attr_type)
@@ -174,6 +197,10 @@ class AttributeDeserializationError(TypeError):
 
 
 class AttributeNullError(ValueError):
+    """
+    Raised when an attribute which is not nullable (:code:`null=False`) is unset during serialization.
+    """
+
     def __init__(self, attr_name: str) -> None:
         self.attr_path = attr_name
 
